import Foundation

/**
Stellt eine Palette auf der Zeichenfläche dar.

- author: Albert Wiedemann
- version: 1.0
*/
class PalettenSymbol
{
    /** x-Koordinate */
    private var x: Int
    /** y-Koordinate */
    private var y: Int
    /** Höhe */
    private let höhe: Int
    /** Farbe */
    private let farbe: String
    /** Gewicht für die Beschriftung */
    private let gewicht: Int
    /** Rahmen */
    private let rahmen: Rechteck
    /** Fläche */
    private let fläche: Rechteck
    /** Text */
    private let text: Text

    /**
    Legt die Elemente des Symbols an
    - parameters:
        - x: x-Koordinate
        - y: y-Koordinate
        - höhe: Höhe der Palette
        - gewicht: Gewicht für die Beschriftung
        - farbe: Farbe
    */
    init(x: Int, y: Int, höhe: Int, gewicht: Int, farbe: String)
    {
        self.gewicht = gewicht
        self.x = x
        self.y = y
        self.höhe = höhe
        self.farbe = farbe
        rahmen = Rechteck()
        rahmen.PositionSetzen(x: x, y: y)
        rahmen.GrößeSetzen(breite: 100, höhe: höhe)
        rahmen.FarbeSetzen(farbe: "schwarz");
        fläche = Rechteck()
        fläche.PositionSetzen(x: x + 2, y: y + 2)
        fläche.GrößeSetzen(breite: 96, höhe: höhe - 4)
        fläche.FarbeSetzen(farbe: farbe)
        text = Text()
        text.PositionSetzen(x: x + 40, y: y + höhe - 4)
        text.TextSetzen(text: "\(gewicht)")
    }
    
    /**
    Erzeugt ein neues Symbol mit gleichen Daten
    - returns: Kopie
    */
    func KopieErstellen() -> PalettenSymbol
    {
        return PalettenSymbol(x: x, y: y, höhe: höhe, gewicht: gewicht, farbe: farbe)
    }
        
    /**
    Setzt die Position (der linken oberen Ecke) des Symbols.
    - parameters:
        - x: x-Position der linken oberen Ecke
        - y: y-Position der linken oberen Ecke
    */
    func PositionSetzen(x: Int, y: Int)
    {
        self.x = x
        self.y = y
        rahmen.PositionSetzen(x: x, y: y)
        fläche.PositionSetzen(x: x + 2, y: y + 2)
        text.PositionSetzen(x: x + 40, y: y + höhe - 4)
    }

    /**
    Setzt die Position (der linken oberen Ecke) des Symbols.
    - parameters:
        - x: x-Position der linken oberen Ecke
    */
    func XPositionSetzen(x: Int)
    {
        self.x = x
        rahmen.PositionSetzen(x: x, y: y)
        fläche.PositionSetzen(x: x + 2, y: y + 2)
        text.PositionSetzen(x: x + 40, y: y + höhe - 4)
    }
    
    /**
    Gibt die Höhe des Symbols zurück.
    - returns: Höhe
    */
    func HöheGeben() -> Int
    {
        return höhe
    }
    
    /**
    Entfernt das Symbol aus der Anzeige
    */
    func Entfernen()
    {
        rahmen.Entfernen()
        fläche.Entfernen()
        text.Entfernen()
    }
}


/**
Stellt einen Pfeil für den Vorratsstapel dar.

- author: Albert Wiedemann
- version: 1.0
*/
class PfeilSymbol
{
    /** Pfeillinie */
    private let linie: Rechteck
    /** Pfeilspitze */
    private let spitze: Dreieck
    /** Beschriftung */
    private let beschriftung: Text

    /**
    Legt das Pfeilsymbol an.
    - parameters:
        - höhe: Anzeigehöhe (Pfeilspitze)
        - tiefe: die anzuzeigende Rekursionstiefe
    */
    init(höhe: Int, tiefe: Int)
    {
        linie = Rechteck()
        linie.PositionSetzen(x: 210, y: höhe)
        linie.GrößeSetzen(breite: 30, höhe: 1)
        linie.FarbeSetzen(farbe: "schwarz")
        spitze = Dreieck()
        spitze.GrößeSetzen(breite: 10, höhe: 10)
        spitze.PositionSetzen(x: 210, y: höhe - 5)
        spitze.FarbeSetzen(farbe: "schwarz")
        spitze.Drehen(grad: 90)
        beschriftung = Text()
        beschriftung.PositionSetzen(x: 245, y: höhe + 4)
        beschriftung.TextSetzen(text: "\(tiefe)")
    }
    
    /**
    Entfernt das Pfeilsymbol aus der Anzeige
    */
    func Entfernen()
    {
        linie.Entfernen()
        spitze.Entfernen()
        beschriftung.Entfernen()
    }
    
    /**
    Setzt das Symbol auf eine neue Höhe
    - parameters:
         - höhe: die neue Höhe
    */
    func HöheSetzen(höhe: Int)
    {
        linie.PositionSetzen(x: 210, y: höhe)
        spitze.PositionSetzen(x: 210, y: höhe - 5)
        beschriftung.PositionSetzen(x: 245, y: höhe + 4)
    }
}

/**
Beschreibt die Daten einer Palette
 
- author: Albert Wiedemann
- version: 1.0
 */
class Palette
{
    /** Höhe der Palette */
    private let höhe: Int
    /** Gewicht der Palette */
    private let gewicht: Int
    
    /**
    Der Konstruktor legt die Daten fest
    - parameters:
        - h: Höhe der Palette
        - g: Gewicht der Palette
    */
    init(h: Int, g: Int)
    {
        höhe = h
        gewicht = g
    }
    
    /**
    Gibt die Höhe der Palette zurück
    - returns: Höhe
     */
    func HöheGeben() -> Int
    {
        return höhe
    }
    
    /**
    Gibt das Gewicht der Palette zurück
    - returns: Gewicht
    */
    func GewichtGeben() -> Int
    {
        return gewicht
    }
}

/**
Verwaltet die Anzeige der Palettenstapel

- author: Albert Wiedemann
- version: 1.0
*/
class Stapelverwaltung: Ereignisbehandlung
{
    /** Die möglichen Palettenfarben */
    private let farben = ["rot", "grün", "blau", "gelb", "cyan", "magenta", "orange", "braun", "weiß"]
    /** Anzeige des Vorratsstapels */
    private var vorratsStapel: [PalettenSymbol]
    /** Anzeige des aktuellen Stapels */
    private var aktuellerStapel: [PalettenSymbol]
    /** Anzeige des optimalen Stapels */
    private var optimalerStapel: [PalettenSymbol]
    /** Feld der Markierungspfeile */
    private var pfeile: [PfeilSymbol]
    /** Tick des Taktgeber da */
    private var tickDa: Bool
    /** Bezeichner des Palettenvorrats */
    private let bezeichnerVorrat: Text
    /** Bezeichner des aktuellen Stapels */
    private let bezeichnerAktuell: Text
    /** Bezeichner des optimalen Stapels */
    private let bezeichnerOptimal: Text
    /** Höhenbegrenzung für einen Palettenstapel */
    private let oberkante: Rechteck
    /** Bezeichner der Oberkante */
    private let bezeichnerOberkante: Text

    /**
     * Legt den Vorratsstapel an
     * @param vorrat der Palettenvorrat
     * @param maximaleHöhe die maximale Höhe, die ein Palettenstapel haben darf
     */
    init(vorrat: [Palette], maximaleHöhe: Int)
    {
        vorratsStapel = [PalettenSymbol]()
        aktuellerStapel = [PalettenSymbol]()
        optimalerStapel = [PalettenSymbol]()
        pfeile = [PfeilSymbol]()
        var farbenIndex = 0
        var aktHöhe = 450
        for p in vorrat
        {
            let h = p.HöheGeben()
            aktHöhe -= h
            vorratsStapel.append(PalettenSymbol(x: 100, y: aktHöhe, höhe: h, gewicht: p.GewichtGeben(), farbe: farben[farbenIndex % farben.count]))
            farbenIndex += 1
        }
        tickDa = false
        bezeichnerVorrat = Text()
        bezeichnerVorrat.TextSetzen(text: "Palettenvorrat")
        bezeichnerVorrat.PositionSetzen(x: 110, y: 470)
        bezeichnerAktuell = Text()
        bezeichnerAktuell.TextSetzen(text: "Aktueller Versuch")
        bezeichnerAktuell.PositionSetzen(x: 400, y: 470)
        bezeichnerOptimal = Text()
        bezeichnerOptimal.TextSetzen(text: "Bisher bester Stapel")
        bezeichnerOptimal.PositionSetzen(x: 600, y: 470)
        oberkante = Rechteck();
        oberkante.GrößeSetzen(breite: 400, höhe: 3);
        oberkante.PositionSetzen(x: 400, y: 450 - maximaleHöhe - 3);
        oberkante.FarbeSetzen(farbe: "schwarz")
        bezeichnerOberkante = Text()
        bezeichnerOberkante.TextSetzen(text: "Maximale Stapelhöhe")
        bezeichnerOberkante.PositionSetzen(x: 400, y: 450 - maximaleHöhe - 20)
        super.init()
        TaktdauerSetzen(dauer: 100)
        Starten()
    }
    
    /**
    Wartet auf den nächsten Tick des Taktgebers
    */
    func AufTaktimpulsWarten()
    {
    }
    
    /**
    Die Aktionsmethode des Zeitgebers.
    */
    override func TaktImpulsAusführen()
    {
        tickDa = true
    }
    
    /**
    Setzt den optimalen Stapel
    
    pre: der aktuelle Stapel ist optimal.
    */
    func OptimalenStapelSetzen()
    {
        for s in optimalerStapel
        {
            s.Entfernen()
        }
        optimalerStapel.removeAll()
        for s in aktuellerStapel
        {
            let neu = s.KopieErstellen()
            neu.XPositionSetzen(x: 600)
            optimalerStapel.append(neu)
        }
        AufTaktimpulsWarten()
    }
    
    /**
    Fügt den Eintrag für die Palette mit dem gegebenen Index zum aktuellen Stapel hinzu
    - parameters:
        - index: der Index der Palette;
    */
    func AktuellHinzufügen(index: Int)
    {
        let neu = vorratsStapel[index].KopieErstellen()
        var h = 0
        for s in aktuellerStapel
        {
            h += s.HöheGeben()
        }
        neu.PositionSetzen(x: 400, y: 450 - (h + neu.HöheGeben()))
        aktuellerStapel.append(neu)
        AufTaktimpulsWarten()
    }
    
    /**
    Entfernt das oberste Element vom aktuellen Stapel.
    */
    func AktuellEntfernen()
    {
        let alt = aktuellerStapel.remove(at: aktuellerStapel.count - 1)
        alt.Entfernen()
        AufTaktimpulsWarten()
    }
    
    /**
    Setzt einen Markierungspfeil auf den Vorratsstapel
    - parameters:
        - index: der Index in den Vorratsstapel
        - tiefe: die anzuzeigende Rekursionstiefe
     */
    func PfeilSetzen(index: Int, tiefe: Int)
    {
        var h = 0
        for i in 0 ..< index
        {
            h += vorratsStapel[i].HöheGeben()
        }
        h += vorratsStapel[index].HöheGeben() / 2
        if tiefe == pfeile.count + 1
        {
            
            pfeile.append(PfeilSymbol(höhe: 450 - h, tiefe: tiefe))
        }
        else
        {
            pfeile[tiefe-1].HöheSetzen(höhe: 450 - h)
        }
        AufTaktimpulsWarten()
    }
    
    /**
    Löscht den obersten Pfeil.
    */
    func PfeilLöschen()
    {
        let alt = pfeile.remove(at: pfeile.count - 1)
        alt.Entfernen()
    }
}

/**
Rekursive Lösung eines Rucksackproblems

- author: Albert Wiedemann
- version: 1.0
*/
class StapelAufgabe
{
    /** Feld der vorhandenen Paletten */
    private var paletten: [Palette]
    /** Feld mit den Palettenindizes der (bisher) optimalen Lösung */
    private var optimaleLösung: [Palette]
    /** (bisher) optimales Gewicht */
    private var optimalesGewicht: Int
    /** maximale Höhe */
    private let maximaleHöhe: Int
    /** Verwaltung der Stapelanzeigen */
    let verwaltung: Stapelverwaltung
    
    /**
    Der Konstruktor erstellt die Paletten und besetzt die Attribute vor
    */
    init()
    {
        paletten = [Palette]()
        paletten.append(Palette(h: 80, g: 9))
        paletten.append(Palette(h: 90, g: 8))
        paletten.append(Palette(h: 50, g: 5))
        paletten.append(Palette(h: 55, g: 6))
        paletten.append(Palette(h: 15, g: 3))
        paletten.append(Palette(h: 30, g: 2))
        paletten.append(Palette(h: 25, g: 4))
        optimalesGewicht = 0
        optimaleLösung = [Palette]()
        maximaleHöhe = 150
        verwaltung = Stapelverwaltung(vorrat: paletten, maximaleHöhe: maximaleHöhe)
    }
    
    /**
    Sucht die Palettenkonfiguration mit der maximalen Höhe
    */
    func MaximumsucheStarten()
    {
        Schritt(aktIndex: 0, aktHöhe: 0, aktGewicht: 0, lösungsansatz: [Palette](), tiefe: 1)
        print("Das maximal erreichbare Gewicht ist: \(optimalesGewicht)")
    }
    
    /**
    Führt den nächsten Schritt der Suche aus
    - parameters:
        - aktIndex: der Index, ab dem die Paletten ausprobiert werden sollen
        - aktHöhe: die aktuelle Höhe des Stapels
        - aktGewicht: das aktuelle Gewicht des Stapels
        - lösungsansatz: Feld mit den bisher auf dem Stapel befindlichen Paletten
        - tiefe die Rekursionstiefe
    */
    private func Schritt(aktIndex: Int, aktHöhe: Int, aktGewicht: Int, lösungsansatz: [Palette], tiefe: Int)
    {
        if aktGewicht > optimalesGewicht
        {
            optimalesGewicht = aktGewicht
            optimaleLösung = lösungsansatz
            // Alle Botschaften an das Objekt verwaltung dienen nur der grafischen Darstellung und
            // sind nicht Teil des Backtrackings
            verwaltung.OptimalenStapelSetzen()
        }
        if aktIndex < paletten.count
        {
            for i in aktIndex ..< paletten.count
            {
                let p = paletten[i]
                verwaltung.PfeilSetzen(index: i, tiefe: tiefe);
                if p.HöheGeben() + aktHöhe <= maximaleHöhe
                {
                    verwaltung.AktuellHinzufügen(index: i);
                    Schritt(aktIndex: i + 1, aktHöhe: p.HöheGeben() + aktHöhe, aktGewicht: p.GewichtGeben() + aktGewicht, lösungsansatz: lösungsansatz + [p], tiefe: tiefe + 1)
                    verwaltung.AktuellEntfernen()
                }
            }
            verwaltung.PfeilLöschen()
        }
    }
}


let s = StapelAufgabe()
s.MaximumsucheStarten()
    
    
  
